// -*- c++ -*-
//
// Copyright (c) 2002-2003 Indiana University.  All rights reserved.
// Copyright (c) 1996, 1997, 1998, 2000 University of Notre Dame.
//                         All rights reserved.
// 
// This file is part of the OOMPI software package.  For license
// information, see the LICENSE file in the top level directory of the
// OOMPI source distribution.
//
// $Id: user_datatypes.cc,v 1.4 2003/03/03 21:57:01 jewillco Exp $
//
// This program demonstrates how to create a user-defined OOMPI_Datatype.
// To accomplish this, OOMPI provides a base class OOMPI_User_type from
// which all non-base type objects that will be communicated must be
// derived.  This class provides an interface to the OOMPI_Message and
// OOMPI_Array_message classes so that objects derived from
// OOMPI_User_type can be send using the same syntax as base classes.

// Besides inheriting from OOMPI_User_type, the user must also construct
// objects in the derived class so that an underlying MPI_Datatype can be
// built.  OOMPI provides a streams=-based interface to make this process
// easier.

#include <iostream>
#include "oompi.h"

using namespace std;

//
// Class Foo
//
class foo : virtual public OOMPI_User_type {
  friend ostream& operator<<(ostream &os, foo &data);

public:

  // Foo constructor: Builds the OOMPI_Datatype.

  foo();
 
  // Primitive functions just for demonstration
 
  void Set(int ii1, int* ii2, float ff1, float ff2, char cc1, 
		  double* dd1) {
    int i;

    i1 = ii1; f1 = ff1; f2 = ff2; c1 = cc1;
    for (i = 0; i < 3; i++)
      i2[i] = ii2[i];
    for (i = 0; i < 5; i++)
      d1[i] = dd1[i];
  }

private:
 
  // The actual data to be communicated
 
  int i1, i2[3];
  float f1, f2;
  char c1;
  double d1[5];
  
  // Static OOMPI_Datatype variable holds the newly
  // constructed MPI_Datatype
  
  static OOMPI_Datatype type;
  
};

 
// The foo constructor in this example first checks to see if 
// the OOMPI_Data type for the class has been written yet.  If not,
// if builds the datatype with the streams interface.  
foo::foo(void)
: OOMPI_User_type(type, this, 201) 
{
  // Check to see if we have already built the datatype (it
  // is a class variable)
  
  if (!type.Built()) {
    type.Struct_start(this);

    // The << operator can be used to stream in scalar variables.
    // The Entry() member function must be used to set array
    // variables, since a count argument must be given.  However,
    // the Entry() member function returns an OOMPI::Datatype,
    // so one can continue streaming in scalar variables.
    
    type << i1;
    type.Entry(i2, 3) << f1 << f2 << c1;
    type.Entry(d1, 5);

    // The Struct_end() member function finished the definition, and
    // performs an MPI_TYPE_COMMIT.  This class can now be sent
    // and received in OOMPI messages.

    type.Struct_end();
  }
};


// Instantiate the class variable type in foo.
OOMPI_Datatype foo::type;


// Simplistic print function for the foo class
ostream &
operator<<(ostream &os, foo& data)
{
  int i;

  os << "i1:" << data.i1 << " ";
  for (i = 0; i < 3; i++)
    os << "i2[" << i << "]:" << data.i2[i] << " ";
  os << endl << "           f1:" << data.f1 << " f2:" << data.f2 
     << " c1:" << data.c1 << endl << "           ";
  for (i = 0; i < 5; i++)
    os << "d1[" << i << "]:" << data.d1[i] << " ";

  return os;
}


// Contrary to what it looks like, this program is extremely simple.
// Each process in OOMPI_COMM_WORLD will send a message of type foo to
// the console process, which displays the message using the
// overloaded << operator.
int
main(int argc, char *argv[])
{
  OOMPI_COMM_WORLD.Init(argc, argv);

  int rank = OOMPI_COMM_WORLD.Rank();
  int size = OOMPI_COMM_WORLD.Size();

  // Instantiate a foo, and build its corresponding OOMPI Datatype

  foo msg;
  
  if (rank == 0) {

    // Receive and display the message from each of the other processes 
    for (int i = 1; i < size; i++) {
      OOMPI_COMM_WORLD[i].Recv(msg);
      cout << "from " << i << " --> " << msg << endl << endl;
    }
  }
  else {

    // Set msg with some bogus values and send it to process 0

    int i, stuff1[3];
    double stuff2[5];

    for (i = 0; i < 3; i++)
      stuff1[i] = rank * (i + 1);
    for (i = 0; i < 5; i++)
      stuff2[i] = rank * (i + 0.1);

    msg.Set(rank, stuff1, ((float) rank) / size, 
	    ((float) size) / rank, (rank%26) + 'A', stuff2);
    OOMPI_COMM_WORLD[0].Send(msg);
  }
  
  OOMPI_COMM_WORLD.Finalize();
  return 0;
}

