// -*- c++ -*-
//
// Copyright (c) 2002-2003 Indiana University.  All rights reserved.
// Copyright (c) 1996, 1997, 1998, 2000 University of Notre Dame.
//                         All rights reserved.
// 
// This file is part of the OOMPI software package.  For license
// information, see the LICENSE file in the top level directory of the
// OOMPI source distribution.
//
// $Id: oompi_test.cc,v 1.17 2003/02/13 21:07:48 jsquyres Exp $
//
// OOMPI test program
// Test *all* OOMPI functionality
//

#include <iostream>

#include <string.h>

#include "oompi.h"
#include "oompi_test.h"


//
// Global variables
//

int comm_size = -1;
int my_rank = -1;
int to = -1;
int from = -1;
const double Epsilon = 0.001;
bool flags[SKIP_MAX];


//
// Local functions
//

static void check_args(int argc, char* argv[]);
static int my_strcasecmp(const char* a, const char* b);
static void check_minimals(void);


//
// main
//
int
main(int argc, char *argv[])
{
  // Start up MPI

  check_args(argc, argv);
  OOMPI_COMM_WORLD.Init(argc, argv);

  // Define some globals

  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  to = (my_rank + 1) % comm_size;
  from = (my_rank + comm_size - 1) % comm_size;

  // Announce

  if (my_rank == 0)
    std::cout << std::endl 
	      << "OOMPI Test suite" << std::endl
	      << "----------------" << std::endl
	      << "OSL Version " << OOMPI_VERSION
	      << std::endl << std::endl
	      << "*** There are delays built into some of the tests" 
	      << std::endl 
	      << "*** Please let them complete" << std::endl
	      << "*** No test should take more than 10 seconds" 
	      << std::endl << std::endl;

  // Check for minimal testing conditions in MPI environment

  check_minimals();

  // Catch all fatal signals

  signal_init();

  // Ensure that all the ranks have the relevant command line args
  // That is, pass on any of the _flag arguments

#if OOMPI_BOOL_NE_INT
  MPI_Bcast(flags, SKIP_MAX * sizeof(bool), MPI_CHAR, 0, MPI_COMM_WORLD);
#else
  MPI_Bcast(flags, SKIP_MAX, MPI_INT, 0, MPI_COMM_WORLD);
#endif

  // Test all the objects

  OOMPI_Comm_world_test();

  OOMPI_Environment_test();

  OOMPI_Op_test();
  OOMPI_Group_test();
  OOMPI_Intra_comm_test();

  OOMPI_Datatype_test();
  OOMPI_Packed_test();

  OOMPI_Tag_test();
  OOMPI_Message_test();
  OOMPI_Array_message_test();

  OOMPI_Port_test();

  OOMPI_Request_test();
  OOMPI_Request_array_test();
  OOMPI_Status_test();
  OOMPI_Status_array_test();

  OOMPI_Cart_comm_test();
  OOMPI_Graph_comm_test();
  OOMPI_Inter_comm_test();

  // All done.  Call MPI_Finalize()

  OOMPI_COMM_WORLD.Finalize();

  if (my_rank == 0)
    std::cout << std::endl 
	      << "OOMPI test suite: All done.  All tests passed." << std::endl;

  return 0;
}


//
// Check the args for command line flags
//
void
check_args(int argc, char *argv[])
{
  int i;

  for (i = 0; i <= SKIP_MAX; i++)
    flags[i] = false;

  for (i = 1; i < argc; i++) {
    if (my_strcasecmp(argv[i], "-cray1104") == 0)
      flags[SKIP_CRAY1104] = true;
    else if (my_strcasecmp(argv[i], "-hpux0102") == 0)
      flags[SKIP_HPUX0102] = true;

    else if (my_strcasecmp(argv[i], "-ibm21014") == 0)
      flags[SKIP_IBM21014] = true;
    else if (my_strcasecmp(argv[i], "-ibm21015") == 0)
      flags[SKIP_IBM21015] = true;
    else if (my_strcasecmp(argv[i], "-ibm21016") == 0)
      flags[SKIP_IBM21016] = true;
    else if (my_strcasecmp(argv[i], "-ibm21017") == 0)
      flags[SKIP_IBM21017] = true;
    else if (my_strcasecmp(argv[i], "-ibm21018") == 0)
      flags[SKIP_IBM21018] = true;
    else if (my_strcasecmp(argv[i], "-ibm23012") == 0)
      flags[SKIP_IBM23012] = true;

    else if (my_strcasecmp(argv[i], "-sgi20") == 0)
      flags[SKIP_SGI20] = true;
    else if (my_strcasecmp(argv[i], "-sgi30") == 0)
      flags[SKIP_SGI30] = true;
    else if (my_strcasecmp(argv[i], "-sgi32") == 0)
      flags[SKIP_SGI32] = true;
    else if (my_strcasecmp(argv[i], "-help") == 0 ||
	     my_strcasecmp(argv[i], "-h") == 0) {
      std::cout << "The following command line options are available:" 
		<< std::endl 
		<< " -help       This message" << std::endl 
		<< " -hpux0102   Skip tests for buggy HP-UX 1.02" << std::endl
		<< " -cray1104   Skip tests for cuggy Cray 1.1.0.4" 
		<< std::endl
		<< " -ibm21014   Skip tests for buggy SP2 MPI 2.1.0.14" 
		<< std::endl
		<< " -ibm21015   Skip tests for buggy SP2 MPI 2.1.0.15" 
		<< std::endl
		<< " -ibm21016   Skip tests for buggy SP2 MPI 2.1.0.16" 
		<< std::endl
		<< " -ibm21017   Skip tests for buggy SP2 MPI 2.1.0.17" 
		<< std::endl
		<< " -ibm21018   Skip tests for buggy SP2 MPI 2.1.0.18" 
		<< std::endl
		<< " -ibm23012   Skip tests for buggy SP2 MPI 2.3.0.12" 
		<< std::endl
		<< " -sgi20      Skip tests for buggy SGI MPI 2.0" 
		<< std::endl
		<< " -sgi30      Skip tests for buggy SGI MPI 3.0 and 3.1" 
		<< std::endl
		<< " -sgi32      Skip tests for buggy SGI MPI 3.2 (only needed in 64-bit mode)" << std::endl;
      exit(0);
    }
  }
}


int
my_strcasecmp(const char* a, const char* b)
{
  while ((a != 0) && (b != 0) && (*a != '\0') && (*b != '\0') && (*a == *b))
    a++, b++;
  if (*a == *b)
    return 0;
  return 1;
}


//
// Do meaningless work to burn up time
//
void
do_work(int top)
{
  double start = MPI_Wtime() + .5;
  top++;

  while(MPI_Wtime() < start)
    continue;
}


//
// Check for minimal MPI environment
//
static void
check_minimals()
{
  if (my_rank == 0)
    std::cout << "Test suite running on " << comm_size << " nodes" 
	      << std::endl << std::endl;

  if (comm_size < 2) {
    if (my_rank == 0) {
      std::cout << "Sorry, the OOMPI test suite must be run with at least 2 processors" << std::endl;
      std::cout << "Please re-run the program with 2 or more processors." 
		<< std::endl << std::endl;
    }
    OOMPI_COMM_WORLD.Finalize();
    exit(1);
  }
  
  if ((comm_size % 2) != 0) {
    if (my_rank == 0)
      std::cout << "The OOMPI test suite can only run on an even number"
		<< std::endl
		<< "of processors.  Please re-run the program with an even" 
		<< std::endl
		<< "number of ranks." << std::endl << std::endl;

    OOMPI_COMM_WORLD.Finalize();
    exit(1);
  }
}


